;;************************************************************************
;; graphobj.lsp 
;; contains code for boxplot, quantile, quantile-quantile  
;; and normal-probability plot objects. 
;; copyright (c) 1993-94 by Forrest W. Young
;;************************************************************************

;;Does not require ViSta system. 

;;************************************************************************
;; boxplot-proto
;;************************************************************************

(defun boxplot (data &key (title "Box, Diamond & Dot Plot") 
                     variable-labels point-labels
                     (boxes t) diamonds median-line mean-line
                     location size (go-away t) (show t))
"Args: (data &key title variable-labels point-labels
             (boxes t) diamonds median-line mean-line
             location size (go-away t) (show t))
DATA is a sequence, a list of sequences or a matrix. Makes parallel box and dot plots or parallel diamond and dot plots.  Plot is of the sequence or sequences in the list or the columns of the matrix.  Box plots are based on the five number summary.  Diamond plots are based on the mean and standard deviation.  Widths of boxes and diamonds are proportional to number of observations in the sequence. Parallel boxes may be connected by a line running through their medians.  Parallel diamonds may be connected by a line running through their means. The BOXES, DIAMONDS, MEAN-LINE and MEDIAN-LINE arguments determine which features are initially displayed. Default is that only boxes are displayed."
  (send boxplot-proto :new data :title title :point-labels point-labels
        :variable-labels variable-labels :location location :boxes boxes
        :diamonds diamonds :median-line median-line :mean-line mean-line
        :size size :go-away go-away :show show))

(defproto boxplot-proto '(data x boxes diamonds mean-line median-line) 
  () scatterplot-proto)

(defmeth boxplot-proto :data (&optional (sequence nil set))
"Args: (&optional sequence)
Sets or returns the plot data for the y-axis."
  (if set (setf (slot-value 'data) sequence))
  (slot-value 'data))

(defmeth boxplot-proto :x (&optional (list nil set))
"Args: (&optional list)
Sets or returns the plot data for the x-axis."
  (if set (setf (slot-value 'x) list))
  (slot-value 'x))

(defmeth boxplot-proto :boxes (&optional (logical nil set))
"Args: (&optional logical)
Sets or returns t or nil for boxes showing."
  (if set (setf (slot-value 'boxes) logical))
  (slot-value 'boxes))

(defmeth boxplot-proto :diamonds (&optional (logical nil set))
"Args: (&optional logical)
Sets or returns t or nil for diamonds showing."
  (if set (setf (slot-value 'diamonds) logical))
  (slot-value 'diamonds))

(defmeth boxplot-proto :mean-line (&optional (logical nil set))
"Args: (&optional logical)
Sets or returns t or nil for mean line showing."
  (if set (setf (slot-value 'mean-line) logical))
  (slot-value 'mean-line))

(defmeth boxplot-proto :median-line (&optional (logical nil set))
"Args: (&optional logical)
Sets or returns t or nil for median line showing."
  (if set (setf (slot-value 'median-line) logical))
  (slot-value 'median-line))

(defmeth boxplot-proto :isnew 
  (data &key title variable-labels point-labels 
        boxes diamonds mean-line median-line location size go-away show)
  (call-next-method 2 :title title :variable-labels variable-labels
                    :location location :size size :go-away go-away :show nil)
  (let* ((n nil)
         (x nil)
         (loc nil))
    (send self :x-axis nil nil 0)
    (send self :new-menu "BoxPlot" 
              :items '(LINK SHOWING-LABELS MOUSE RESIZE-BRUSH DASH 
                            ERASE-SELECTION FOCUS-ON-SELECTION SHOW-ALL 
                            SYMBOL COLOR SELECTION))
    (send self :margin 0 17 0 0)
    (send self :add-overlay (send boxplot-overlay-proto :new))
    (send self :boxes boxes)
    (send self :diamonds diamonds)
    (send self :median-line median-line)
    (send self :mean-line mean-line)
    (send self :new-plot data 
          :title title
          :variable-labels variable-labels
          :point-labels point-labels)
    (when show (send self :show-window))
    self))

(defmeth boxplot-proto :new-plot 
  (data &key title variable-labels point-labels) 
  (let ((n nil)
        (x nil)
        (loc nil))
    (send self :start-buffering)
    (setq data 
          (cond ((matrixp data) (column-list data))
                ((or (not (listp data)) (numberp (car data))) (list data))
                (t data)))
    (send self :clear)
    (send self :data data)
    (setf n (mapcar #'length data))
    (setf x (* (mean n) (iseq (length n))))
    (send self :x x)
    (when (not variable-labels) 
          (setf variable-labels (repeat "" (length data))))
    (send self :variable-labels variable-labels)
    (let ((range (get-nice-range (min data) (max data) 4)))
      (send self :range 1 (nth 0 range) (nth 1 range))
      (send self :y-axis t (second (send self :y-axis)) (nth 2 range)))
    (send self :range 0 (- (first x) (/ (first n) 2))
          (+ (last  x) (/ (last  n) 2)))
   
    (dotimes (i (length data))
             (send self :add-points 
                   (repeat (nth i x) (length (nth i data))) (nth i data)))
    (when point-labels
          (send self :point-label (iseq (length point-labels)) point-labels))
    (send self :change-plot)
    (send self :buffer-to-screen)))
  

(defmeth boxplot-proto :redraw ()
  (send self :start-buffering)
  (call-next-method)
  (let* ((data (send self :data))
         (x (send self :x))
         (n (mapcar #'length data))
         (variable-labels (send self :variable-labels))
         (w 0)
         (loc nil))
    (dotimes (i (length data))
             (if (or (send self :diamonds) (send self :boxes))
                 (setf w (/ (nth i n) 4))
                 (setf w 0))
             (setf loc (send self :scaled-to-canvas 
                             (- (nth i x) w) (mean (nth i data))))
             (send self :draw-text-up (select variable-labels i) 
                   (- (first loc) 8) (second loc) 1 0)))
  (send self :buffer-to-screen))

(defmeth boxplot-proto :add-diamondplot 
  (y &key (x 1.0) (width 1.0) (draw t))
  (unless (= 2 (send self :num-variables)) (error "only works for 2D plots"))
  (let* ((half-box (* 0.5 width))
         (half-foot (* 0.1 width))
         (mean (mean y))
         (stdv (standard-deviation y))
         (low (min y))
         (p1 (- mean (if stdv stdv 0)))
         (p3 (+ mean (if stdv stdv 0)))
         (high (max y)))
    (send self :add-lines (list (- x half-box) (+ x half-box))
                          (list mean mean))
    (send self :add-lines (list x (- x half-box) x (+ x half-box) x)
                          (list p3 mean p1 mean p3))))

(defmeth boxplot-proto :add-boxplot (y &key (x 1.0) (width 1.0) (draw t))
  (unless (= 2 (send self :num-variables)) (error "only works for 2D plots"))
  (let* ((half-box (* 0.4 width))
         (half-foot (* 0.30 width))
         (fiv (fivnum y))
         (low (quantile y .10))
         (q1 (select fiv 1))
         (med (select fiv 2))
         (q3 (select fiv 3))
         (high (quantile y .90)))
    (send self :add-lines (list (- x half-foot) (+ x half-foot))
                          (list low low))
    (send self :add-lines (list (- x half-box) (+ x half-box))
                          (list med med))
    (send self :add-lines (list (- x half-foot) (+ x half-foot))
                          (list high high))
    (send self :add-lines (list (- x half-box) (- x half-box) 
                                (+ x half-box) (+ x half-box) (- x half-box))
                          (list q3 q1 q1 q3 q3))))

(defmeth boxplot-proto :add-median-line ()
  (send self :add-lines (send self :x) (mapcar #'median (send self :data))))

(defmeth boxplot-proto :add-mean-line ()
  (send self :add-lines (send self :x) (mapcar #'mean (send self :data))))

(defmeth boxplot-proto :switch-boxes ()
  (send self :boxes (not (send self :boxes)))
  (send self :change-plot))

(defmeth boxplot-proto :switch-diamonds ()
  (send self :diamonds (not (send self :diamonds)))
  (send self :change-plot))

(defmeth boxplot-proto :switch-mean-line ()
  (send self :mean-line (not (send self :mean-line)))
  (send self :change-plot))

(defmeth boxplot-proto :switch-median-line ()
  (send self :median-line (not (send self :median-line)))
  (send self :change-plot))

(defmeth boxplot-proto :change-plot ()
  (let* ((data (send self :data))
         (n (mapcar #'length data))
         (x (send self :x)))
    (send self :start-buffering)
    (send self :clear-lines)
    (dotimes (i (length data))
             (when (send self :boxes)
                   (send self :add-boxplot (nth i data) 
                         :width (/ (nth i n) 2) :x (nth i x)))
             (when (send self :diamonds)
                   (send self :add-diamondplot (nth i data) 
                         :width (/ (nth i n) 2) :x (nth i x))))
    (when (send self :median-line) 
          (send self :add-lines x (mapcar #'median data)))
    (when (send self :mean-line)
          (send self :add-lines x (mapcar #'mean data)))
    (send self :redraw)
    (send self :buffer-to-screen)
    ))

(defproto boxplot-overlay-proto '() () graph-overlay-proto)

(defmeth boxplot-overlay-proto :redraw ()
  (let* ((graph (send self :graph))
         (twBoxes (send graph :text-width "Boxes"))
         (twDiamonds (send graph :text-width "Diamonds"))
         (twMedianLine (send graph :text-width "Median Line")))
    (send self :draw-button (send graph :boxes) 10 3 10 10)
    (send graph :draw-string "Boxes" 22 12)
    (send self :draw-button (send graph :diamonds) (+ 30 twboxes) 3 10 10)
    (send graph :draw-string "Diamonds" (+ 42 twboxes) 12)
    (send self :draw-button 
          (send graph :median-line) (+ 50 twboxes twdiamonds) 3 10 10)
    (send graph :draw-string "Median Line" (+ 62 twboxes twdiamonds) 12)
    (send self :draw-button
          (send graph :mean-line) 
          (+ 70 twboxes twdiamonds twmedianline) 3 10 10)
    (send graph :draw-string "Mean Line" 
          (+ 82 twboxes twdiamonds twmedianline) 12)
    (send graph :draw-line 0 16 (send graph :canvas-width) 16)))

(defmeth boxplot-overlay-proto :do-click (x y m1 m2)
  (let* ((graph (send self :graph))
         (twBoxes (send graph :text-width "Boxes"))
         (twDiamonds (send graph :text-width "Diamonds"))
         (twMedianLine (send graph :text-width "Median Line")))
    (when (< 3 y 13)
          (cond
            ((< 10 x 20) 
             (send self :draw-button (not (send graph :boxes)) 10 3 10 10)
             (send graph :switch-boxes))
            ((< (+ 30 twboxes) x (+ 40 twboxes)) 
             (send self :draw-button 
                   (not (send graph :diamonds)) (+ 30 twboxes) 3 10 10)
             (send graph :switch-diamonds))
            ((< (+ 50 twboxes twdiamonds) x (+ 60 twboxes twdiamonds))
             (send self :draw-button (not (send graph :median-line))
                   (+ 50 twboxes twdiamonds) 3 10 10)
             (send graph :switch-median-line))
            ((< (+ 70 twboxes twdiamonds twmedianline) x
                (+ 80 twboxes twdiamonds twmedianline))
             (send self :draw-button (not (send graph :mean-line))
                   (+ 70 twboxes twdiamonds twmedianline) 3 10 10)
             (send graph :switch-mean-line))))))

(defmeth boxplot-overlay-proto :draw-button (paint a b c d)
  (when paint (send (send self :graph) :paint-rect a b c d))
  (when (not paint)
        (send (send self :graph) :draw-color 'white)
        (send (send self :graph) :paint-rect a b c d)
        (send (send self :graph) :draw-color 'black)
        (send (send self :graph) :frame-rect a b c d)))

;;************************************************************************
;; quantile-plot-proto
;;************************************************************************

(defun quantile-plot (data &key reg-line (title "Quantile Plot") nice-range
                           (variable-label "Values") (points nil) point-labels 
                           location size (go-away t) (show t))
"Args: (data &key reg-line (title Quantile Plot) nice-range (variable-label value) (points nil) point-labels location size (go-away t) (show t))
DATA is a sequence. Makes quantile plot of the sequence. By default a line plot is produced without a menu. Points and a menu are added if POINTS is T. When REG-LINE is t a regression line is to be added to the plot.  Range and tick-marks for y are determined by get-nice-range unless NICE-RANGE (list of min, max, num tick marks) is specified. Remaining arguments are the same as for scatterplots. Function written by Forrest Young to replace original by Luke Tierney."
  (send quantile-plot-proto :new data :title title :point-labels point-labels
        :variable-label variable-label :location location :size size
        :show-points-and-menu points 
        :reg-line reg-line :nice-range nice-range
        :go-away go-away :show show))

(defun qplot (data &rest args)
"Alias for quantile-plot"
  (apply #'quantile-plot data args))

(defproto quantile-plot-proto '(data x reg-line show-points-and-menu) () 
  scatterplot-proto)

(defmeth quantile-plot-proto :data (&optional (sequence nil set))
"Args: (&optional sequence)
Sets or returns the plot data for the y-axis."
  (if set (setf (slot-value 'data) sequence))
  (slot-value 'data))

(defmeth quantile-plot-proto :x (&optional (list nil set))
"Args: (&optional list)
Sets or returns the plot data for the x-axis."
  (if set (setf (slot-value 'x) list))
  (slot-value 'x))

(defmeth quantile-plot-proto :reg-line (&optional (logical nil set))
"Args: (&optional list)
Sets or returns t or nil for whether there is a regression line."
  (if set (setf (slot-value 'reg-line) logical))
  (slot-value 'reg-line))

(defmeth quantile-plot-proto :show-points-and-menu 
  (&optional (logical nil set))
"Args: (&optional list)
Sets or returns t or nil for whether to show points and menu."
  (if set (setf (slot-value 'show-points-and-menu) logical))
  (slot-value 'show-points-and-menu))

(defmeth quantile-plot-proto :isnew 
  (data &key title variable-label point-labels reg-line nice-range
        show-points-and-menu
        location size go-away show)
  (call-next-method 2 :title title :variable-label variable-label
                    :location location :size size :go-away go-away :show nil)
  (send self :show-points-and-menu show-points-and-menu)
  (if show-points-and-menu
      (send (send self :menu) :title "QPlot")
      (send self :menu nil))
  (send self :range 0 0 1)
  (send self :x-axis t t 5)
  (send self :new-plot data 
        :variable-label variable-label
        :point-labels point-labels
        :reg-line reg-line
        :nice-range nice-range)
  (when show (send self :show-window))
  self)

(defmeth quantile-plot-proto :new-plot 
  (data &key variable-label point-labels reg-line nice-range)
  (let* ((index (order data))
         (y (sort-data data))
         (x (sort-data (/ (1+ (rank y)) (1+ (length y)))))
         (ab nil)
         (a nil)
         (b nil))
    (send self :start-buffering)
    (send self :clear)
    (send self :data data)
    (send self :x x)
    (let ((y-range nice-range))
      (when (not y-range)
            (setf y-range (get-nice-range (min data) (max data) 4)))
      (send self :range 1 (nth 0 y-range) (nth 1 y-range))
      (send self :y-axis t t (nth 2 y-range)))
    (when (> (length x) 1) (send self :add-lines x y))
    (when (send self :show-points-and-menu)
          (send self :add-points (select x index) (select y index)))
    (when point-labels
          (send self :point-label (iseq (length point-labels)) point-labels))
    (when (and reg-line (> (length (remove-duplicates x)) 1))
          (setf ab (send (regression-model x y :print nil)
                         :coef-estimates))
          (setf a (first ab))
          (setf b (second ab))
          (send self :add-lines 
                (list (/ (- (min y) a) b) (/ (- (max y) a) b))
                (list (min y) (max y))))
    (send self :variable-label 0 "Quantiles")
    (when (not variable-label) (setf variable-label ""))
    (send self :variable-label 1 variable-label)
    (send self :redraw)
    (send self :buffer-to-screen)
    ))

;;************************************************************************
;; quantile-quantile-plot-proto
;;************************************************************************

(defun quantile-quantile-plot 
  (xdata ydata &key reg-line (title "Quantile-Quantile Plot") 
     nice-x-range nice-y-range (points nil)
     (variable-labels (list "Ordered X Values" "Ordered Y Values"))  
     point-labels location size (go-away t) (show t))
"Args (xdata ydata &key points (title Quantile-Quantile Plot) reg-line point-labels  variable-labels location size (go-away t) (show t))
XDATA and YDATA are sequences which do not have to be the same length. Makes a quantile-quantile plot of the two sequences.  By default a line plot is produced without a menu. Points and a menu are added if POINTS is T. When REG-LINE is t a regression line is to be added to the plot.  Range and tick-marks for x and y are determined by get-nice-range unless NICE-RANGE-X and NICE-RANGE-Y (lists of min, max, num tick marks) are specified. Remaining arguments are the same as for scatterplots. Function written by Forrest Young to replace original by Luke Tierney."
  (send quantile-quantile-plot-proto :new xdata ydata :reg-line reg-line 
        :nice-x-range nice-x-range :nice-y-range nice-y-range
        :title title :point-labels point-labels
        :show-points-and-menu points
        :variable-labels variable-labels :location location 
        :size size :go-away go-away :show show))

(defun qqplot (xdata ydata &rest args)
  (apply #'quantile-quantile-plot xdata ydata args))

(defproto quantile-quantile-plot-proto 
  '(xdata ydata reg-line show-points-and-menu) () scatterplot-proto)

(defmeth quantile-quantile-plot-proto :xdata (&optional (list nil set))
"Args: (&optional list)
Sets or returns the plot data for the x-axis."
  (if set (setf (slot-value 'xdata) list))
  (slot-value 'xdata))

(defmeth quantile-quantile-plot-proto :ydata (&optional (sequence nil set))
"Args: (&optional sequence)
Sets or returns the plot data for the y-axis."
  (if set (setf (slot-value 'ydata) sequence))
  (slot-value 'ydata))

(defmeth quantile-quantile-plot-proto :reg-line (&optional (logical nil set))
"Args: (&optional list)
Sets or returns t or nil for whether there is a regression line."
  (if set (setf (slot-value 'reg-line) logical))
  (slot-value 'reg-line))

(defmeth quantile-quantile-plot-proto :show-points-and-menu 
  (&optional (logical nil set))
"Args: (&optional list)
Sets or returns t or nil for whether to show points and menu."
  (if set (setf (slot-value 'show-points-and-menu) logical))
  (slot-value 'show-points-and-menu))

(defmeth quantile-quantile-plot-proto :isnew 
  (xdata ydata &key reg-line nice-x-range nice-y-range title variable-labels 
     show-points-and-menu point-labels location size go-away show)
  (call-next-method 2 :title title 
                    :location location :size size :go-away go-away :show nil)
  (send self :show-points-and-menu show-points-and-menu)
  (if show-points-and-menu
      (send (send self :menu) :title "QQPlot")
      (send self :menu nil))
  (send self :new-plot xdata ydata
        :variable-labels variable-labels
        :point-labels point-labels
        :reg-line reg-line
        :nice-x-range nice-x-range
        :nice-y-range nice-y-range)
  (when show (send self :show-window))
  self)

(defmeth quantile-quantile-plot-proto :new-plot 
  (xdata ydata &key variable-labels point-labels reg-line 
         nice-x-range nice-y-range)
  (let* ((x-sort (if (= (length xdata) (length ydata))
                     (sort-data xdata)
                     (if (= (length ydata) 1)
                         (list (mean xdata))
                         (quantile xdata (rseq 0 1 (length ydata))))))
         (y-sort (if (= (length xdata) (length ydata))
                     (sort-data ydata)
                     (if (= (length ydata) 1)
                         ydata
                         (quantile ydata (rseq 0 1 (length ydata)))))))
    (send self :start-buffering)
    (send self :clear)
    (send self :xdata x-sort)
    (send self :ydata y-sort)
    (let ((x-range nice-x-range)
          (y-range nice-y-range))
      (when (not x-range)
            (setf x-range (get-nice-range (min xdata) (max xdata) 4)))
      (send self :range 0 (nth 0 x-range) (nth 1 x-range))
      (send self :x-axis t t (nth 2 x-range))
      (when (not y-range)
            (setf y-range (get-nice-range (min ydata) (max ydata) 4)))
      (send self :range 1 (nth 0 y-range) (nth 1 y-range))
      (send self :y-axis t t (nth 2 y-range)))
    (when (send self :show-points-and-menu)
          (send self :add-points x-sort y-sort))
    (when (> (length x-sort) 1) (send self :add-lines x-sort y-sort))
    (when (and reg-line (> (length (remove-duplicates x-sort)) 1))
          (let* ((ab (send (regression-model x-sort y-sort :print nil) 
                           :coef-estimates))
                 (a (first ab))
                 (b (second ab)))
            (send self :add-lines 
                  (list (/ (- (min ydata) a) b) (/ (- (max ydata) a) b))
                  (list (min ydata) (max ydata)))))
    (when (not variable-labels) (setf variable-labels (list "" "")))
    (when (not (first  variable-labels)) (setf (first  variable-labels) ""))
    (when (not (second variable-labels)) (setf (second variable-labels) ""))
    (send self :variable-label 0 (first variable-labels))
    (send self :variable-label 1 (second variable-labels))
    (send self :redraw)
    (send self :buffer-to-screen)
    ))

;;************************************************************************
;; normal probability plot method
;;************************************************************************

(defun npplot (data &rest args)
  (apply #'normal-probability-plot data args))

(defun normal-probability-plot (data &key (title "Normal Probability Plot")
                                  point-labels
                                  (variable-label "Value")
                                  location
                                  size
                                  (show t))
"Args: (data &key (title Normal Probability Plot) point-labels
       (variable-label Value) location size (show t))
Makes a normal probability plot for DATA, using title, point-labels and
variable-label as supplied."
     (let* ((seq (iseq (length data)))
            (x (/ (- seq (mean seq)) (standard-deviation seq)))
            (y (sort-data data))
            (ab (send (regression-model x y :print nil) :coef-estimates))
            (npplot (plot-points x y
                     :title title
                     :variable-labels (list "Normal Scores" variable-label)
                     :point-labels point-labels
                     :location location
                     :size size
                     :show nil)))
       (send npplot :abline (first ab) (second ab))
       (send npplot :add-lines x y)
       (when show (send npplot :show-window))
       npplot))
