;;########################################################################
;; mmrmob2.lsp
;; multivariate multiple regression ViSta model object
;; Copyright (c) 1991-95 by Forrest W. Young
;; Multivariate Test implemented by Mark G. Baxter
;; Redundancy Analysis implemented by Mark G. Baxter and Forrest W. Young 
;;########################################################################

(require "mmrmob1")

(defmeth mmr-model-object-proto :analysis ()
  (let* ((dv (send self :dv))
         (iv (send self :iv))
         (intercept (send self :intercept))
         (weights (send self :weights))
         (all-coefs nil)
         (all-coefs-mat nil)
         (all-scores nil)
         (all-scores-mat nil)
         (rm ())
         (num-dv (length dv))
         (num-iv (length iv))
         (numobs (send self :nobs)) 
         )
    (when (>= num-iv numobs) 
          (error (format nil "Cannot analyze data that does not have more observations (~d) than predictor variables (~d)." numobs num-iv)))
    (when (< (send self :redundancy) 0)
          (error-message "The minimum number of redundancy variates is zero. Redundancy analysis will not be performed."))
    (when (> (send self :redundancy) (min num-dv num-iv))
          (send self :redundancy (min num-dv num-iv))
          (error-message (format nil "The number of redundancy variates has been lowered to the maximum of ~d." (min num-dv num-iv))))
        
    (when (<= (send self :redundancy) 0) (send self :redundancy nil))
 
    (dolist (i dv)
            (setf rm (concatenate 'list rm (list  
               (regression-model 
                (col (send self :data-matrix) iv)
                (col (send self :data-matrix) i )
                :intercept intercept
                :predictor-names (select (send self :variables) iv)
                :response-name   (select (send self :variables) i )
                :weights (if weights 
                             (col (send self :data-matrix) weights)
                             nil)
                :print nil
                )))))
    (dotimes (i num-dv)
             (setf all-coefs (combine all-coefs 
                    (if intercept 
                       (rest (send (select rm i) :coef-estimates))
                    (send (select rm i) :coef-estimates))))
             (setf all-scores 
                 (combine all-scores 
                          (send (select rm i) :fit-values)))
             )
    (setf all-coefs  (rest all-coefs))
    (setf all-scores (rest all-scores))
    (setf all-coefs-mat (matrix (list num-dv num-iv) all-coefs))
    (setf all-scores-mat 
               (transpose (matrix (list num-dv numobs) all-scores)))
    (send self :num-iv num-iv)
    (send self :num-dv num-dv)
    (send self :coefs all-coefs-mat)
    (send self :scores all-scores-mat)
    (send self :reg-models rm)
    (send self :compute-cor-cov)
    (send self :y-stdv (sqrt (mapcar #'variance (column-list 
                       (send self :get-y-matrix)))))
    (send self :x-stdv (sqrt (mapcar #'variance (column-list 
                       (send (first (send self :reg-models)) :x)))))
    (send self :beta   (matmult (diagonal (send self :x-stdv)) 
                       (transpose (send self :coefs)) 
                       (inverse (diagonal (send self :y-stdv)))))
    (if (and (> (send self :num-dv) 1) (send self :redundancy))
        (send self :compute-redundancy-variables))
    ))

(defmeth mmr-model-object-proto :compute-cor-cov ()
  (let* ((x-mat   (send (first (send self :reg-models)) :x))
         (y-mat   (send self :get-y-matrix))
         (yx-mat  (bind-columns y-mat x-mat))) 
    (send self :cor (correlation-matrix yx-mat))
    (send self :cov (covariance-matrix  yx-mat))))

(defmeth mmr-model-object-proto :get-y-matrix ()
    (let* ((y-mat (send (first (send self :reg-models)) :y))
           (y-mat-2 nil))
      (dotimes (i (1- (send self :num-dv)))
               (setf y-mat-2 
                     (bind-rows y-mat (send (select (send self :reg-models) 
                                                  (1+ i)) :y)))
               (setf y-mat y-mat-2))
      (if (= 1 (send self :num-dv))
          (matrix (list (length y-mat) 1) y-mat)
          (transpose y-mat))))

(defmeth mmr-model-object-proto :get-intercept-vector ()
  (let* ((int-vec (first (send (first (send self :reg-models)) 
                               :coef-estimates)))
         (int-vec-2 nil))
    (dotimes (i (1- (send self :num-dv)))
             (setf int-vec-2 
                   (combine int-vec (first (send (select 
                            (send current-model :reg-models) (1+ i)) 
                                                 :coef-estimates))))
             (setf int-vec int-vec-2))
    int-vec))

(defmeth mmr-model-object-proto :multivariate-test ()
    (let* ((x-mat (send (first (send self :reg-models)) :x))
           (x-mat-1 (bind-columns (repeat 1 (first (array-dimensions x-mat))) 
                                  x-mat))
           (y-mat (send self :get-y-matrix))
           (beta (bind-rows (send self :get-intercept-vector) 
                            (transpose (send self :coefs))))
           (c (transpose (bind-columns 
                          (repeat 0 (second (array-dimensions x-mat)))
                          (diagonal 
                           (repeat 1 
                                   (second (array-dimensions x-mat)))))))
           (s-h (matmult (transpose (matmult (transpose c) beta)) 
                         (inverse (matmult (transpose c) 
                                           (inverse 
                                            (matmult (transpose x-mat-1)
                                                     x-mat-1)) c)) 
                         (transpose c) beta))
           (resid (- y-mat (matmult x-mat-1 beta)))
           (s-e (matmult (transpose resid) resid))
           (wl
            (cond
              ((> (determinant (+ s-h s-e)) 1E-9)
               (/ (determinant s-e) (determinant (+ s-h s-e))))
              (t 1)))
           (p (array-dimension (+ s-h s-e) 0))
           (q (array-dimension 
               (matmult (transpose c) 
                (inverse (matmult (transpose x-mat-1) x-mat-1))
                c) 0))
           (v (- (1- (array-dimension x-mat 0)) (array-dimension x-mat 1)))
           (s (min (list p q)))
           (m (* 0.5 (1- (abs (- p q)))))
           (n (* 0.5 (1- (- v p))))
           (r (- v (/ (+ (- p q) 1) 2)))
           (u (/ (- (* p q) 2) 4))
           (tee (cond
                ((> (- (+ (^ p 2) (^ q 2)) 5) 0)
                 (^ (/ (- (* p p q q) 4) (- (+ (^ p 2) (^ q 2)) 5)) 0.5))
                (t 1)))
           (wl-alt (^ wl (/ 1 tee)))
           (f-stat (/ (* (- 1 wl-alt) (- (* r tee) (* 2 u))) 
                   (* wl-alt (* p q))))
           (num-df (* p q))
           (den-df (- (* r tee) (* 2 u)))
           (p-level (- 1 (f-cdf f-stat num-df den-df)))
           (singflag t)
           (exact (cond
                    ((> (min (list p q)) 2) nil)
                    (t t))))
      (when
       (< (determinant (matmult (transpose x-mat) x-mat)) 1E-9)
       (setf singflag nil))
      (when
       (< (determinant (matmult (transpose y-mat) y-mat)) 1E-9)
       (setf singflag nil))       
       (list wl f-stat num-df den-df p-level exact singflag)))

(defmeth mmr-model-object-proto :compute-redundancy-variables () 
  (let* ((x-mat   (send (first (send self :reg-models)) :x))
         (y-mat   (send self :get-y-matrix))
         (x-stdv  (sqrt (mapcar #'variance (column-list x-mat))))
         (y-stdv  (sqrt (mapcar #'variance (column-list y-mat))))
         (x-means (mapcar (function mean) (column-list x-mat)))
         (y-means (mapcar (function mean) (column-list y-mat)))
         (nx      (array-dimension x-mat 1))
         (ny      (array-dimension y-mat 1))
         (nr      (send self :redundancy))
         (nvar    (array-dimension x-mat 0))
         (x-devia (- x-mat (matrix (list nvar nx) (repeat x-means nvar))))
         (r       (send self :cor))
         (results (send self :redundancy-solve r ny nx nr)))
    (send self :redun-index (first results))
    (send self :redun-evals (second results))
    (send self :redun-evecs (third results))
    (send self :redun-coefs (fourth results))
    (send self :redun-bmat  
          (transpose (matmult (diagonal y-stdv) 
                              (transpose (fifth results))
                              (inverse (diagonal x-stdv)))))
    (send self :redun-scores 
          (+ (matmult x-devia (send self :redun-bmat)) 
             (matrix (list nvar ny) (repeat y-means nvar))))
    ))

(defmeth mmr-model-object-proto :redundancy-solve (s ny nx nr) 
"Args: S NY NX
Computes the redundancy solution for S, a covariance matrix for NY responses (the first portion of S) and NX predictors (the last portion), using the first NR redundancy variates. Returns a list containing the redundancy index value, eigen-values, eigen-vectors, A-matrix (redundancy variate coefficients) and Beta-Matrix (standardized redundancy coefficients based on NR variates)."
  (let* ((nrank (min ny nx))
         (nall  (+ nx ny))
         (iy    (iseq ny))
         (ix    (iseq ny (1- nall)))
         (s-xx  (select s ix ix))
         (s-yx  (select s iy ix))
         (s-xy  (transpose s-yx)) 
         (prod  (matmult s-yx (inverse s-xx) (transpose s-yx)))
         (e-sol (eigen prod))
         (evals (select (first e-sol) (iseq nrank)))
         (redun-index (/ (sum evals) ny))
         (r-di  (sqrt (inverse (diagonal evals))))
         (evecs (second e-sol))
         (u-d   (transpose (matrix (list nrank ny) 
                                   (combine (select evecs (iseq nrank))))))  
         (a     (matmult      (inverse s-xx) s-xy u-d r-di));jackson
        ;(a     (matmult r-di (inverse s-xx) s-xy u-d     ));tyler
         (beta  (matmult (select a (iseq nx) (iseq nr)) 
                         (transpose (select a (iseq nx) (iseq nr))) 
                         s-xy)))
    (list redun-index evals u-d a beta)))
    
(defmeth mmr-model-object-proto :visualize ()
  (if (not (eq current-object self)) (setcm self))
  (when (= 1 (send self :num-iv))
        (error "Visualization not possible when there is only one predictor. Use univariate Regression Analysis."))
  (let* ((cpoints (center (select (send self :data-matrix)
                                  (iseq (send self :nobs))
                                  (send self :iv))))
         (rays (transpose (send self :beta)));(rays (send self :coefs))
         (nredun (send self :redundancy))
         (redun-rays (if nredun
          (matmult 
           (diagonal  (select (send self :redun-evals) (iseq nredun)))
           (transpose (select (send self :redun-coefs)
                              (iseq (send self :num-iv)) (iseq nredun))))))
         (numerator (max (mapcar #'ssq (row-list cpoints))))
         (denom (max (mapcar #'ssq (row-list rays))))
         (scale-factor (sqrt (/ numerator denom)))
         
         (spin-dim (min 9 (send self :num-iv)))
         (scaled-coefs (* scale-factor
                          (select (column-list rays) (iseq spin-dim))))
         (redlabels nil)
         (denom-redun nil)
         (scale-factor-redun nil)
         (scaled-redun-coefs nil)
         (ray-labels (select (send self :variables) (send self :dv))))
    (when redun-rays
          (setf redlabels
                (mapcar #'(lambda (x) (format nil "Red~d" x))
                        (iseq nredun)))
          (setf denom-redun 
                (max (mapcar #'ssq (append (row-list rays)
                                           (row-list redun-rays)))))
          (setf scale-factor-redun (sqrt (/ numerator denom-redun)))
          (setf scaled-redun-coefs 
                (* scale-factor-redun
                   (select (column-list redun-rays) (iseq spin-dim)))))
    (when (> (send self :num-iv) 2)
          (spreadplot (column-list cpoints)
                      (send self :labels) 
                      (select (send self :variables) (send self :iv))
                      :scale-type 'fixed
                      :show-histogram nil
                      :show-reg-plot t)
          (send spin-plot   :depth-cuing nil)
          #+color(send spin-plot   :point-color 
                       (iseq (send spin-plot :num-points)) 'green)
          (send spin-plot   :point-symbol
                (iseq (send spin-plot :num-points)) 'square)
          (send spin-plot   :add-rays scaled-coefs 
                :ray-labels ray-labels :ray-color 'blue)
          (when redun-rays 
                (send spin-plot :add-rays scaled-redun-coefs 
                      :ray-labels redlabels :ray-color 'red))
          (send list-obs :title "Labels")
          (send spin-plot :redraw)
          (setf reg-plot 
                (plot-points (list 1 2) (list 1 2)
                             :show nil :location location22 :size graph-size))
          (send reg-plot :add-slot 'switch 1)
          (send reg-plot :title "Regression Plot")
          (defmeth reg-plot :close ()
            (send scatmat :close))
          (defmeth reg-plot :reg-plot (i mob) 
            (let* ((x (send (select (send mob :reg-models) i) :y))
                   (y (col (send mob :scores) i))
                   (min (min (combine x y)))
                   (max (max (combine x y)))
                   (gnr (get-nice-range min max 5))
                   (axmin (first gnr))
                   (axmax (second gnr))
                   )
              (send reg-plot :range 0 (first gnr) (second gnr) :draw nil)
              (send reg-plot :range 1 (first gnr) (second gnr) :draw nil)
              (send reg-plot :x-axis t t (third gnr) :draw nil)
              (send reg-plot :y-axis t t (third gnr) :draw nil)

              (send reg-plot :clear)
              (send reg-plot :add-points x y 
                    :point-labels (send mob :labels) :draw nil)
              (send reg-plot :variable-label 0
                    (select (select (send mob :variables) 
                                    (send mob :dv)) i))
              (send reg-plot :variable-label 1 
                    (strcat "Predicted " 
                            (select (select (send mob :variables) 
                                            (send mob :dv)) i)))
            
              (send reg-plot :add-lines 
                    (list (list axmin axmax) (list axmin axmax)) :draw nil)
            
              (send reg-plot :redraw)))

          (defmeth reg-plot :adjust-screen-point (i)
            (let* ((npoints (send self :num-points)))
              (send self :slot-value 'switch 
                    (* -1 (send self :slot-value 'switch)))
              (cond
                ((> npoints i) (call-next-method i))
                (t 
                 (when (and (< i (+ npoints (send current-model :num-dv)))
                            (> 0 (send self :slot-value 'switch)))
                       (send self :reg-plot (- i npoints) current-model))
                 ))))
          (send reg-plot :reg-plot 0 self)
          (send reg-plot :show-window)
          (send reg-plot :showing-labels t)
          (send reg-plot :linked t)
          #+color(send reg-plot :use-color t)
          #+color(send reg-plot :point-color 
                       (iseq (send reg-plot :num-points)) 'green)
          (send reg-plot :point-symbol
                (iseq (send reg-plot :num-points)) 'square)
          )
    (when (= (send self :num-iv) 2)
          (setf scatterplot (plot-points 
                    (column-list cpoints)
                    :title (send self :title)
                    :location location11
                    :size (+ (/ menu-bar-height 2) (* 2 graph-size))
                    :point-labels (send self :labels)
                    :variable-labels 
                    (select (send self :variables) (send self :iv))
                    :show nil))
          (send scatterplot :showing-labels t)
          (send scatterplot :linked t)
          (send scatterplot :new-menu "ScatPlot" 
              :items '(LINK SHOWING-LABELS MOUSE RESIZE-BRUSH DASH 
                            ERASE-SELECTION FOCUS-ON-SELECTION SHOW-ALL 
                            SYMBOL COLOR SELECTION))
          (send scatterplot :redraw)
          (send scatterplot :show-window)
          (defmeth scatterplot :close ()
            (send list-obs :remove)
            (call-next-method))
          (setf list-obs (send current-object :obs-namelist))
          (send list-obs  :linked t)
          (defmeth list-obs :close ()
            (send scatterplot :close))
          (send list-obs :new-menu "Labels" 
                :items '(LINK MOUSE DASH ERASE-SELECTION FOCUS-ON-SELECTION 
                              SHOW-ALL COLOR SELECTION))
          )
    #+color(send scatterplot :use-color t)
    #+color(send scatterplot :point-color 
                 (iseq (send scatterplot :num-points)) 'green)
    (send scatterplot :point-symbol
          (iseq (send scatterplot :num-points)) 'square)
    (send scatterplot :add-rays scaled-coefs 
          :ray-labels ray-labels :ray-color 'blue)
    (send list-obs :add-points (send self :num-dv) 
          :point-labels ray-labels)
    (when redun-rays 
          (send scatterplot :add-rays scaled-redun-coefs 
                :ray-labels redlabels :ray-color 'red)
          (send list-obs :add-points (send self :redundancy)
                      :point-labels redlabels))
    (send scatterplot :redraw)
    t))

(provide "mmrmob2")
