;;########################################################################
;; sprdplt1.lsp
;; Copyright (c) 1996 by Forrest W. Young, Richard A. Faldowski, & Carla Bann
;; SpreadPlot Object
;;########################################################################

;
;Spread-Plot constructor function
;

(defun spread-plot 
  (plot-matrix &key model menu-title)
"Method Args: PLOTMATRIX &KEY CREATING-OBJECT MENU-TITLE
PLOTMATRIX is a matrix of plot-objects. The spreadplot cells consist of plots identified in this matrix. CREATING-OBJECT is an optional object identification of the model or data object that is creating the spreadplot (nil, the default, when there is none). MENU-TITLE specifies the title of the spreadplots menu (nil, the default, means there is no menu). Creates and returns a spreadplot supervisor object."
  (let ((spobj (send spreadplot-proto :new 
                     :model model 
                     :menu-title menu-title))) 
    (send spobj :plot-matrix plot-matrix)
    (send spobj :create-spreadplot)
    (when menu-title (send spobj :create-menu menu-title))
    (send spobj :show-spreadplot)
    spobj))

;
;Define prototype spreadplot and its isnew method
;

(defproto spreadplot-proto 
  '(model plot-matrix menu  menu-title menu-template))

(defmeth spreadplot-proto :isnew 
  (&key model 
        menu-title 
        (menu-template '(dash dash dash)) )
  (if model (setf (slot-value 'model) model))
  (if menu-title (setf (slot-value 'menu-title) menu-title))
  (if menu-template (setf (slot-value 'menu-template) menu-template)) )

;
;Slot Accessor Methods
;

(defmeth spreadplot-proto :model (&optional (values nil set))
"Method Args:  &optional (VALUES nil) 
Sets or retrieves the object id# of the model in which the spreadplot is installed."
  (if set (setf (slot-value 'model) values))
  (slot-value 'model) )

(defmeth spreadplot-proto :plot-matrix (&optional (matrix nil set))
"Method Args:  &optional PLOT-MATRIX
Sets or retrieves the spreadplot's matrix of plot object ids."
  (if set (setf (slot-value 'plot-matrix) matrix))
  (slot-value 'plot-matrix) )

(defmeth spreadplot-proto :menu (&optional (values nil set))
"Method Args:  &optional MENU
Sets or retrieves the spreadplot's menu in its MENU slot."
  (if set (setf (slot-value 'menu) values))
  (slot-value 'menu) )

(defmeth spreadplot-proto :menu-title (&optional (values nil set))
"Method Args:  &optional ''TITLE''
Sets or retrieves the title of the spreadplot's menu."
  (if set (setf (slot-value 'menu-title) values))
  (slot-value 'menu-title) )

(defmeth spreadplot-proto :menu-template (&optional (values nil set))
"Method Args:  &optional TEMPLATE
Sets or retrieves the template of items for the spreadplot's menu."
  (if set (setf (slot-value 'menu-template) values))
  (slot-value 'menu-template) )

;
; Additional Methods
;

(defmeth spreadplot-proto :create-spreadplot ()
"Method Args: ()
Creates the layout for the spreadplot cells and displays the plots."
  (let* ((plot-matrix (send self :plot-matrix))
         (nrows (first (size plot-matrix)))
         (ncols (second (size plot-matrix)))
         (plot-size nil)
         (location (make-array (list nrows ncols)))
         (plot nil)
         (plot00 (aref plot-matrix 0 0))
         )
    (when (< nrows 1) (error "SpreadPlots must have at least one row"))
    (when (< ncols 1) (error "SpreadPlots must have at least one column"))
    (setf plot-size  
          (min (- (floor (/ (first screen-size) ncols))
                  window-decoration-width 2)
               (- (floor (/ (- (second screen-size) menu-bar-height) nrows))
                  msdos-fiddle
                  (/ window-decoration-height 2))))
    (dotimes (i nrows)
         (dotimes (j ncols)
              (setf (aref location i j)
                    (list (* j (+ plot-size window-decoration-width 2))
                          (+ (* i (+ plot-size 2)) 
                             (* msdos-fiddle (+ 1 i))
                             (* (1+ i) window-decoration-height))))))
    (dotimes (i nrows)
       (dotimes (j ncols)
          (setf plot (aref plot-matrix i j))
          (defmeth plot :close () (send plot00 :close))
          (send plot :size (+ plot-size window-decoration-width) plot-size)
          (apply #'send plot :location (aref location i j))
          (send plot :show-window)))
    (send plot00 :show-window)
    (defmeth plot00 :close ()
      (dotimes (i nrows)
               (dotimes (j ncols)
                        (send (aref plot-matrix i j) :remove))))
    ))

(defmeth spreadplot-proto :remove ()
"Method Args: () 
Removes the spreadplot.  First, remove its menu if it has one.  Then, remove the individual plots installed in it. "
  (when (send self :menu)
        (send (send self :menu) :remove)
        (send (send self :menu) :dispose) )
  (mapcar #'(lambda (plot) 
              (when plot
                    (if (send plot :allocated-p)
                        (send plot :remove))))
              (send self :all-plots)) )

(defmeth spreadplot-proto :show-spreadplot ()
"Method Args: () 
Shows the spreadplot.  First, shows the individual plots installed in the spreadplot, then installs the menu (if it has one) into the menu bar."
  (send self :show-visible-plots)
  (when (send self :menu)
        (send (send self :menu) :install) ))

(defmeth spreadplot-proto :hide-spreadplot ()
"Method Args: () 
Hides the spreadplot.  First, removes its menu (if it has one) from the menu bar.  Then, hides the individual plots installed in it. "
  (when (send self :menu)
        (send (send self :menu) :remove) )
  (send self :hide-all-plots) )

(defmeth spreadplot-proto :create-menu 
  (&optional title  &key (items (send self :menu-template)) ) 
  (unless title 
          (setq title (slot-value 'menu-title)) )
  (if (slot-value 'menu) 
      (send (slot-value 'menu) :dispose) )
  (flet (
         (make-item (item) (send self :make-menu-item item) )
         )
    (let (
          (menu (send menu-proto :new title) )
          )
      (send self :menu menu) 
      (apply #'send menu :append-items  
             (remove nil (mapcar #'make-item items)))
      menu) ) )

(defmeth spreadplot-proto :make-menu-item (item-template)
  (if (kind-of-p item-template menu-item-proto)
      item-template
      (case item-template
        ( dash (send dash-item-proto :new) ) 
        ( show-plots 
          (send graph-item-proto :new "Show Plots" self 
                :show-visible-plots) )
        ( hide-plots 
          (send graph-item-proto :new "Hide Plots" self 
                :hide-all-plots) )
        ( show-spreadplot 
          (send graph-item-proto :new "Show SpreadPlot" self 
                :show-ssp) )
        ( hide-spreadplot 
          (send graph-item-proto :new "Hide SpreadPlot" self 
                :hide) )
        ( kill-spreadplot
          (send graph-item-proto :new "Kill SpreadPlot" (slot-value 'model)
                :delete-spreadplot-supervisor) ))))

(defmeth spreadplot-proto :all-plots ()
"Method Args: () 
Retrieves the list of a spreadplot-object's own slots and checks whether they are plots or not.  Returns a list of all plots from the object's slots. 

Note: expects any slot to contain one and only one plot. 
" 
(let ((slots (mapcar #'(lambda (x)
                         (send self :slot-value x))
                     (send self :own-slots))))
  (select slots (which (mapcar #'(lambda (x) 
                                   (kind-of-p x graph-proto))
                               slots))) ))

(defmeth spreadplot-proto :hide-all-plots ()
  (mapcar #'(lambda (x)
              (send x :hide-window))
          (send self :all-plots)))

(defmeth spreadplot-proto :show-visible-plots ()
  (mapcar #'(lambda (x)
              (send x :show-window))
          (send self :all-plots)))




